@@ -6,10 +6,12 @@ Created on 2020-08-19 by hbldh <henrik.blidh@nedomkull.com>
 """
 
 import asyncio
+import functools
 import logging
 import sys
 import uuid
 import warnings
+from ctypes import pythonapi
 from typing import Any, Dict, List, Optional, Sequence, Union, cast
 
 import async_timeout
@@ -44,11 +46,15 @@ from bleak_winrt.windows.devices.enumeration import (
     DevicePairingResultStatus,
     DeviceUnpairingResultStatus,
 )
-from bleak_winrt.windows.foundation import EventRegistrationToken
+from bleak_winrt.windows.foundation import (
+    AsyncStatus,
+    EventRegistrationToken,
+    IAsyncOperation,
+)
 from bleak_winrt.windows.storage.streams import Buffer
 
 from ... import BleakScanner
-from ...exc import PROTOCOL_ERROR_CODES, BleakError, BleakDeviceNotFoundError
+from ...exc import PROTOCOL_ERROR_CODES, BleakDeviceNotFoundError, BleakError
 from ..characteristic import BleakGATTCharacteristic
 from ..client import BaseBleakClient, NotifyCallback
 from ..device import BLEDevice
@@ -250,7 +256,7 @@ class BleakClientWinRT(BaseBleakClient):
 
         def handle_services_changed():
             if not self._services_changed_events:
-                logger.warn("%s: unhandled services changed event", self.address)
+                logger.warning("%s: unhandled services changed event", self.address)
             else:
                 for event in self._services_changed_events:
                     event.set()
@@ -360,18 +366,30 @@ class BleakClientWinRT(BaseBleakClient):
             self._session.maintain_connection = True
             # This keeps the device connected until we set maintain_connection = False.
 
-            # wait for the session to become active
-            async with async_timeout.timeout(timeout):
-                await event.wait()
+            wait_connect_task = asyncio.create_task(event.wait())
+
+            try:
+                # a connection may not be made until we request info from the
+                # device, so we have to get services before the GATT session
+                # is set to active
+                wait_get_services_task = asyncio.create_task(self.get_services())
+
+                try:
+                    # wait for the session to become active
+                    async with async_timeout.timeout(timeout):
+                        await asyncio.gather(wait_connect_task, wait_get_services_task)
+
+                finally:
+                    wait_get_services_task.cancel()
+            finally:
+                wait_connect_task.cancel()
+
         except BaseException:
             handle_disconnect()
             raise
         finally:
             self._session_active_events.remove(event)
 
-        # Obtain services, which also leads to connection being established.
-        await self.get_services()
-
         return True
 
     async def disconnect(self) -> bool:
@@ -534,8 +552,6 @@ class BleakClientWinRT(BaseBleakClient):
            A :py:class:`bleak.backends.service.BleakGATTServiceCollection` with this device's services tree.
 
         """
-        if not self.is_connected:
-            raise BleakError("Not connected")
 
         # Return the Service Collection.
         if self._services_resolved:
@@ -568,7 +584,10 @@ class BleakClientWinRT(BaseBleakClient):
                 services_changed_event.wait()
             )
             self._services_changed_events.append(services_changed_event)
-            get_services_task = self._requester.get_gatt_services_async(*args)
+
+            get_services_task = FutureLike(
+                self._requester.get_gatt_services_async(*args)
+            )
 
             try:
                 await asyncio.wait(
@@ -587,10 +606,10 @@ class BleakClientWinRT(BaseBleakClient):
                 "%s: restarting get services due to services changed event",
                 self.address,
             )
-            args = [BluetoothCacheMode.UNCACHED]
+            args = [BluetoothCacheMode.CACHED]
 
         services: Sequence[GattDeviceService] = _ensure_success(
-            get_services_task.get_results(),
+            get_services_task.result(),
             "services",
             "Could not get GATT services",
         )
@@ -885,3 +904,64 @@ class BleakClientWinRT(BaseBleakClient):
 
         event_handler_token = self._notification_callbacks.pop(characteristic.handle)
         characteristic.obj.remove_value_changed(event_handler_token)
+
+
+class FutureLike:
+    """
+    Wraps a WinRT IAsyncOperation in a "future-like" object so that it can
+    be passed to Python APIs.
+
+    Needed until https://github.com/pywinrt/pywinrt/issues/14
+    """
+
+    _asyncio_future_blocking = True
+
+    def __init__(self, async_result: IAsyncOperation) -> None:
+        self._async_result = async_result
+        self._callbacks = []
+        self._loop = asyncio.get_running_loop()
+
+        def call_callbacks(op: IAsyncOperation, status: AsyncStatus):
+            for c in self._callbacks:
+                c(self)
+
+        async_result.completed = functools.partial(
+            self._loop.call_soon_threadsafe, call_callbacks
+        )
+
+    def result(self) -> Any:
+        return self._async_result.get_results()
+
+    def done(self) -> bool:
+        return self._async_result.status != AsyncStatus.STARTED
+
+    def cancelled(self) -> bool:
+        return self._async_result.status == AsyncStatus.CANCELED
+
+    def add_done_callback(self, callback, *, context=None) -> None:
+        self._callbacks.append(callback)
+
+    def remove_done_callback(self, callback) -> None:
+        self._callbacks.remove(callback)
+
+    def cancel(self, msg=None) -> bool:
+        if self._async_result.status != AsyncStatus.STARTED:
+            return False
+        self._async_result.cancel()
+        return True
+
+    def exception(self) -> Optional[Exception]:
+        if self._async_result.status == AsyncStatus.STARTED:
+            raise asyncio.InvalidStateError
+        if self._async_result.status == AsyncStatus.COMPLETED:
+            return None
+        if self._async_result.status == AsyncStatus.CANCELED:
+            raise asyncio.CancelledError
+        if self._async_result.status == AsyncStatus.ERROR:
+            try:
+                pythonapi.PyErr_SetFromWindowsErr(self._async_result.error_code)
+            except OSError as e:
+                return e
+
+    def get_loop(self) -> asyncio.AbstractEventLoop:
+        return self._loop
